@@ -1,8 +1,8 @@
-{-# LANGUAGE ScopedTypeVariables, ScopedTypeVariables #-}
+{-# LANGUAGE ScopedTypeVariables, ScopedTypeVariables, TupleSections #-}
 
 module Happstack.Server.Internal.Handler
     ( request
-    , parseResponse 
+    , parseResponse
     , putRequest
     ) where
 
@@ -32,6 +32,7 @@ import Happstack.Server.Internal.MessageWrap
 import Happstack.Server.SURI(SURI(..),path,query)
 import Happstack.Server.SURI.ParseURI
 import Happstack.Server.Internal.TimeoutIO (TimeoutIO(..))
+import Happstack.Server.Internal.Monads (failResponse)
 import qualified Happstack.Server.Internal.TimeoutManager as TM
 import Numeric
 import System.Directory (removeFile)
@@ -83,7 +84,8 @@ rloop timeoutIO mlog host handler inputStr
 
                      let ioseq act = act >>= \x -> x `seq` return x
 
-                     res <- ioseq (handler req) `E.catch` \(e::E.SomeException) -> return $ result 500 $ "Server error: " ++ show e
+                     (res, handlerKilled) <- ((, False) `liftM` ioseq (handler req))
+                         `E.catch` \(e::E.SomeException) -> return (failResponse (show e), fromException e == Just ThreadKilled)
 
                      case mlog of
                        Nothing -> return ()
@@ -102,11 +104,13 @@ rloop timeoutIO mlog host handler inputStr
                      putAugmentedResult timeoutIO req res
                      -- clean up tmp files
                      cleanupTempFiles req
-                     when (continueHTTP req res) $ rloop timeoutIO mlog host handler nextRequest
+                     -- do not continue if handler was killed
+                     when (not handlerKilled && continueHTTP req res) $
+                         rloop timeoutIO mlog host handler nextRequest
 
 -- NOTE: if someone took the inputs and never put them back, then they are responsible for the cleanup
 cleanupTempFiles :: Request -> IO ()
-cleanupTempFiles req = 
+cleanupTempFiles req =
     do mInputs <- tryTakeMVar (rqInputsBody req)
        case mInputs of
          Nothing -> return ()
